package imports

type Config struct {
	RequireSingleImport bool // Require the use of a single 'import' declaration only
	RequireGrouping     bool // Require the use of grouped 'import' declarations
}
