//go:generate go run ../generate.go

// Package simple contains analyzes that simplify code.
// All suggestions made by these analyzes are intended to result in objectively simpler code,
// and following their advice is recommended.
package simple
