// Code generated by "stringer -output=zipproto_strings.go -type=IPProto -trimprefix=IPProto"; DO NOT EDIT.

package wf

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[IPProtoICMP-1]
	_ = x[IPProtoICMPV6-58]
	_ = x[IPProtoTCP-6]
	_ = x[IPProtoUDP-17]
}

const (
	_IPProto_name_0 = "ICMP"
	_IPProto_name_1 = "TCP"
	_IPProto_name_2 = "UDP"
	_IPProto_name_3 = "ICMPV6"
)

func (i IPProto) String() string {
	switch {
	case i == 1:
		return _IPProto_name_0
	case i == 6:
		return _IPProto_name_1
	case i == 17:
		return _IPProto_name_2
	case i == 58:
		return _IPProto_name_3
	default:
		return "IPProto(" + strconv.FormatInt(int64(i), 10) + ")"
	}
}
