// Code generated by "stringer -output=zaction_strings.go -type=Action -trimprefix=Action"; DO NOT EDIT.

package wf

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[ActionBlock-4097]
	_ = x[ActionPermit-4098]
	_ = x[ActionCalloutTerminating-20483]
	_ = x[ActionCalloutInspection-24580]
	_ = x[ActionCalloutUnknown-16389]
}

const (
	_Action_name_0 = "BlockPermit"
	_Action_name_1 = "CalloutUnknown"
	_Action_name_2 = "CalloutTerminating"
	_Action_name_3 = "CalloutInspection"
)

var (
	_Action_index_0 = [...]uint8{0, 5, 11}
)

func (i Action) String() string {
	switch {
	case 4097 <= i && i <= 4098:
		i -= 4097
		return _Action_name_0[_Action_index_0[i]:_Action_index_0[i+1]]
	case i == 16389:
		return _Action_name_1
	case i == 20483:
		return _Action_name_2
	case i == 24580:
		return _Action_name_3
	default:
		return "Action(" + strconv.FormatInt(int64(i), 10) + ")"
	}
}
