// Copyright (c) Tailscale Inc & AUTHORS
// SPDX-License-Identifier: BSD-3-Clause

//go:build windows

// Package automation provides essential types for interacting with COM Automation (IDispatch).
package automation
