package testify

const (
	ModulePath = "github.com/stretchr/testify"

	AssertPkgName  = "assert"
	HTTPPkgName    = "http"
	MockPkgName    = "mock"
	RequirePkgName = "require"
	SuitePkgName   = "suite"

	AssertPkgPath  = ModulePath + "/" + AssertPkgName
	HTTPPkgPath    = ModulePath + "/" + HTTPPkgName
	MockPkgPath    = ModulePath + "/" + MockPkgName
	RequirePkgPath = ModulePath + "/" + RequirePkgName
	SuitePkgPath   = ModulePath + "/" + SuitePkgName
)
