// Code generated by "stringer -output=zsublayerflags_strings.go -type=fwpmSublayerFlags -trimprefix=fwpmSublayerFlags"; DO NOT EDIT.

package wf

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[fwpmSublayerFlagsPersistent-1]
}

const _fwpmSublayerFlags_name = "Persistent"

var _fwpmSublayerFlags_index = [...]uint8{0, 10}

func (i fwpmSublayerFlags) String() string {
	i -= 1
	if i >= fwpmSublayerFlags(len(_fwpmSublayerFlags_index)-1) {
		return "fwpmSublayerFlags(" + strconv.FormatInt(int64(i+1), 10) + ")"
	}
	return _fwpmSublayerFlags_name[_fwpmSublayerFlags_index[i]:_fwpmSublayerFlags_index[i+1]]
}
