// Code generated by generate.go. DO NOT EDIT.

package quickfix

import (
	"honnef.co/go/tools/analysis/lint"
	"honnef.co/go/tools/quickfix/qf1001"
	"honnef.co/go/tools/quickfix/qf1002"
	"honnef.co/go/tools/quickfix/qf1003"
	"honnef.co/go/tools/quickfix/qf1004"
	"honnef.co/go/tools/quickfix/qf1005"
	"honnef.co/go/tools/quickfix/qf1006"
	"honnef.co/go/tools/quickfix/qf1007"
	"honnef.co/go/tools/quickfix/qf1008"
	"honnef.co/go/tools/quickfix/qf1009"
	"honnef.co/go/tools/quickfix/qf1010"
	"honnef.co/go/tools/quickfix/qf1011"
	"honnef.co/go/tools/quickfix/qf1012"
)

var Analyzers = []*lint.Analyzer{
	qf1001.SCAnalyzer,
	qf1002.SCAnalyzer,
	qf1003.SCAnalyzer,
	qf1004.SCAnalyzer,
	qf1005.SCAnalyzer,
	qf1006.SCAnalyzer,
	qf1007.SCAnalyzer,
	qf1008.SCAnalyzer,
	qf1009.SCAnalyzer,
	qf1010.SCAnalyzer,
	qf1011.SCAnalyzer,
	qf1012.SCAnalyzer,
}
