// Package ifelse provides helpers for analysing the control flow in if-else chains,
// presently used by the following rules:
// - early-return
// - indent-error-flow
// - superfluous-else
package ifelse
