// Code generated using std.sh; DO NOT EDIT.

// We will ignore that fact that some of packages
// were removed from stdlib.

package analyzer

var std = map[string]struct{}{
	// added in Go v1.2 in compare to v1.1 (docker image)
	"archive/tar":         {},
	"archive/zip":         {},
	"bufio":               {},
	"bytes":               {},
	"cmd/cgo":             {},
	"cmd/fix":             {},
	"cmd/go":              {},
	"cmd/gofmt":           {},
	"cmd/yacc":            {},
	"compress/bzip2":      {},
	"compress/flate":      {},
	"compress/gzip":       {},
	"compress/lzw":        {},
	"compress/zlib":       {},
	"container/heap":      {},
	"container/list":      {},
	"container/ring":      {},
	"crypto":              {},
	"crypto/aes":          {},
	"crypto/cipher":       {},
	"crypto/des":          {},
	"crypto/dsa":          {},
	"crypto/ecdsa":        {},
	"crypto/elliptic":     {},
	"crypto/hmac":         {},
	"crypto/md5":          {},
	"crypto/rand":         {},
	"crypto/rc4":          {},
	"crypto/rsa":          {},
	"crypto/sha1":         {},
	"crypto/sha256":       {},
	"crypto/sha512":       {},
	"crypto/subtle":       {},
	"crypto/tls":          {},
	"crypto/x509":         {},
	"crypto/x509/pkix":    {},
	"database/sql":        {},
	"database/sql/driver": {},
	"debug/dwarf":         {},
	"debug/elf":           {},
	"debug/gosym":         {},
	"debug/macho":         {},
	"debug/pe":            {},
	"encoding":            {},
	"encoding/ascii85":    {},
	"encoding/asn1":       {},
	"encoding/base32":     {},
	"encoding/base64":     {},
	"encoding/binary":     {},
	"encoding/csv":        {},
	"encoding/gob":        {},
	"encoding/hex":        {},
	"encoding/json":       {},
	"encoding/pem":        {},
	"encoding/xml":        {},
	"errors":              {},
	"expvar":              {},
	"flag":                {},
	"fmt":                 {},
	"go/ast":              {},
	"go/build":            {},
	"go/doc":              {},
	"go/format":           {},
	"go/parser":           {},
	"go/printer":          {},
	"go/scanner":          {},
	"go/token":            {},
	"hash":                {},
	"hash/adler32":        {},
	"hash/crc32":          {},
	"hash/crc64":          {},
	"hash/fnv":            {},
	"html":                {},
	"html/template":       {},
	"image":               {},
	"image/color":         {},
	"image/color/palette": {},
	"image/draw":          {},
	"image/gif":           {},
	"image/jpeg":          {},
	"image/png":           {},
	"index/suffixarray":   {},
	"io":                  {},
	"io/ioutil":           {},
	"log":                 {},
	"log/syslog":          {},
	"math":                {},
	"math/big":            {},
	"math/cmplx":          {},
	"math/rand":           {},
	"mime":                {},
	"mime/multipart":      {},
	"net":                 {},
	"net/http":            {},
	"net/http/cgi":        {},
	"net/http/cookiejar":  {},
	"net/http/fcgi":       {},
	"net/http/httptest":   {},
	"net/http/httputil":   {},
	"net/http/pprof":      {},
	"net/mail":            {},
	"net/rpc":             {},
	"net/rpc/jsonrpc":     {},
	"net/smtp":            {},
	"net/textproto":       {},
	"net/url":             {},
	"os":                  {},
	"os/exec":             {},
	"os/signal":           {},
	"os/user":             {},
	"path":                {},
	"path/filepath":       {},
	"reflect":             {},
	"regexp":              {},
	"regexp/syntax":       {},
	"runtime":             {},
	"runtime/cgo":         {},
	"runtime/debug":       {},
	"runtime/pprof":       {},
	"runtime/race":        {},
	"sort":                {},
	"strconv":             {},
	"strings":             {},
	"sync":                {},
	"sync/atomic":         {},
	"syscall":             {},
	"testing":             {},
	"testing/iotest":      {},
	"testing/quick":       {},
	"text/scanner":        {},
	"text/tabwriter":      {},
	"text/template":       {},
	"text/template/parse": {},
	"time":                {},
	"unicode":             {},
	"unicode/utf16":       {},
	"unicode/utf8":        {},
	"unsafe":              {},
	// added in Go v1.3 in compare to v1.2 (docker image)
	"cmd/addr2line":  {},
	"cmd/nm":         {},
	"cmd/objdump":    {},
	"cmd/pack":       {},
	"debug/plan9obj": {},
	// added in Go v1.4 in compare to v1.3 (docker image)
	"cmd/pprof": {},
	// added in Go v1.5 in compare to v1.4 (docker image)
	"go/constant":          {},
	"go/importer":          {},
	"go/types":             {},
	"mime/quotedprintable": {},
	"runtime/trace":        {},
	// added in Go v1.6 in compare to v1.5 (docker image)
	// added in Go v1.7 in compare to v1.6 (docker image)
	"context":            {},
	"net/http/httptrace": {},
	// added in Go v1.8 in compare to v1.7 (docker image)
	"plugin": {},
	// added in Go v1.9 in compare to v1.8 (docker image)
	"math/bits": {},
	// added in Go v1.10 in compare to v1.9 (docker image)
	// added in Go v1.11 in compare to v1.10 (docker image)
	// added in Go v1.12 in compare to v1.11 (docker image)
	// added in Go v1.13 in compare to v1.12 (docker image)
	"crypto/ed25519": {},
	// added in Go v1.14 in compare to v1.13 (docker image)
	"hash/maphash": {},
	// added in Go v1.15 in compare to v1.14 (docker image)
	"time/tzdata": {},
	// added in Go v1.16 in compare to v1.15 (docker image)
	"embed":               {},
	"go/build/constraint": {},
	"io/fs":               {},
	"runtime/metrics":     {},
	"testing/fstest":      {},
	// added in Go v1.17 in compare to v1.16 (docker image)
	// added in Go v1.18 in compare to v1.17 (docker image)
	"debug/buildinfo": {},
	"net/netip":       {},
	// added in Go v1.19 in compare to v1.18 (docker image)
	"go/doc/comment": {},
	// added in Go v1.20 in compare to v1.19 (docker image)
	"crypto/ecdh":      {},
	"runtime/coverage": {},
	// added in Go v1.21 in compare to v1.20 (docker image)
	"cmp":              {},
	"log/slog":         {},
	"maps":             {},
	"slices":           {},
	"testing/slogtest": {},
	// added in Go v1.22 in compare to v1.21 (docker image)
	"go/version":   {},
	"math/rand/v2": {},
}
