// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Code generated by generate.go. DO NOT EDIT.

package stdlib

type pkginfo struct {
	name string
	deps string // list of indices of dependencies, as varint-encoded deltas
}

var deps = [...]pkginfo{
	{"archive/tar", "\x03n\x03E<\x01\n\x01$\x01\x01\x02\x05\b\x02\x01\x02\x02\f"},
	{"archive/zip", "\x02\x04d\a\x03\x12\x021<\x01+\x05\x01\x0f\x03\x02\x0e\x04"},
	{"bufio", "\x03n\x84\x01D\x14"},
	{"bytes", "q*Z\x03\fG\x02\x02"},
	{"cmp", ""},
	{"compress/bzip2", "\x02\x02\xf1\x01A"},
	{"compress/flate", "\x02o\x03\x81\x01\f\x033\x01\x03"},
	{"compress/gzip", "\x02\x04d\a\x03\x14mT"},
	{"compress/lzw", "\x02o\x03\x81\x01"},
	{"compress/zlib", "\x02\x04d\a\x03\x12\x01n"},
	{"container/heap", "\xb7\x02"},
	{"container/list", ""},
	{"container/ring", ""},
	{"context", "q[o\x01\r"},
	{"crypto", "\x86\x01oC"},
	{"crypto/aes", "\x10\n\t\x95\x02"},
	{"crypto/cipher", "\x03 \x01\x01\x1f\x11\x1c+Y"},
	{"crypto/des", "\x10\x15\x1f-+\x9c\x01\x03"},
	{"crypto/dsa", "D\x04)\x84\x01\r"},
	{"crypto/ecdh", "\x03\v\f\x10\x04\x16\x04\r\x1c\x84\x01"},
	{"crypto/ecdsa", "\x0e\x05\x03\x04\x01\x10\a\v\x06\x01\x04\f\x01\x1c\x84\x01\r\x05K\x01"},
	{"crypto/ed25519", "\x0e\x1e\x11\a\n\a\x1c\x84\x01C"},
	{"crypto/elliptic", "2?\x84\x01\r9"},
	{"crypto/fips140", "\"\x05"},
	{"crypto/hkdf", "/\x14\x01-\x15"},
	{"crypto/hmac", "\x1a\x16\x13\x01\x111"},
	{"crypto/internal/boring", "\x0e\x02\ri"},
	{"crypto/internal/boring/bbig", "\x1a\xe8\x01M"},
	{"crypto/internal/boring/bcache", "\xbc\x02\x13"},
	{"crypto/internal/boring/sig", ""},
	{"crypto/internal/constanttime", ""},
	{"crypto/internal/cryptotest", "\x03\r\n\b%\x0e\x19\x06\x12\x12 \x04\x06\t\x18\x01\x11\x11\x1b\x01\a\x05\b\x03\x05\v"},
	{"crypto/internal/entropy", "I"},
	{"crypto/internal/entropy/v1.0.0", "B/\x93\x018\x13"},
	{"crypto/internal/fips140", "A0\xbd\x01\v\x16"},
	{"crypto/internal/fips140/aes", "\x03\x1f\x03\x02\x13\x05\x01\x01\x06*\x93\x014"},
	{"crypto/internal/fips140/aes/gcm", "\"\x01\x02\x02\x02\x11\x05\x01\a*\x90\x01"},
	{"crypto/internal/fips140/alias", "\xcf\x02"},
	{"crypto/internal/fips140/bigmod", "'\x18\x01\a*\x93\x01"},
	{"crypto/internal/fips140/check", "\"\x0e\x06\t\x02\xb4\x01Z"},
	{"crypto/internal/fips140/check/checktest", "'\x87\x02!"},
	{"crypto/internal/fips140/drbg", "\x03\x1e\x01\x01\x04\x13\x05\t\x01(\x84\x01\x0f7\x01"},
	{"crypto/internal/fips140/ecdh", "\x03\x1f\x05\x02\t\r2\x84\x01\x0f7"},
	{"crypto/internal/fips140/ecdsa", "\x03\x1f\x04\x01\x02\a\x02\x069\x15oF"},
	{"crypto/internal/fips140/ed25519", "\x03\x1f\x05\x02\x04\v9\xc7\x01\x03"},
	{"crypto/internal/fips140/edwards25519", "\x1e\t\a\x112\x93\x017"},
	{"crypto/internal/fips140/edwards25519/field", "'\x13\x052\x93\x01"},
	{"crypto/internal/fips140/hkdf", "\x03\x1f\x05\t\x06;\x15"},
	{"crypto/internal/fips140/hmac", "\x03\x1f\x14\x01\x019\x15"},
	{"crypto/internal/fips140/mlkem", "\x03\x1f\x05\x02\x0e\x03\x052\xca\x01"},
	{"crypto/internal/fips140/nistec", "\x1e\t\f\f2\x93\x01*\r\x14"},
	{"crypto/internal/fips140/nistec/fiat", "'\x137\x93\x01"},
	{"crypto/internal/fips140/pbkdf2", "\x03\x1f\x05\t\x06;\x15"},
	{"crypto/internal/fips140/rsa", "\x03\x1b\x04\x04\x01\x02\r\x01\x01\x027\x15oF"},
	{"crypto/internal/fips140/sha256", "\x03\x1f\x1d\x01\a*\x15~"},
	{"crypto/internal/fips140/sha3", "\x03\x1f\x18\x05\x011\x93\x01K"},
	{"crypto/internal/fips140/sha512", "\x03\x1f\x1d\x01\a*\x15~"},
	{"crypto/internal/fips140/ssh", "'_"},
	{"crypto/internal/fips140/subtle", "\x1e\a\x1a\xc5\x01"},
	{"crypto/internal/fips140/tls12", "\x03\x1f\x05\t\x06\x029\x15"},
	{"crypto/internal/fips140/tls13", "\x03\x1f\x05\b\a\t2\x15"},
	{"crypto/internal/fips140cache", "\xae\x02\r&"},
	{"crypto/internal/fips140deps", ""},
	{"crypto/internal/fips140deps/byteorder", "\x9c\x01"},
	{"crypto/internal/fips140deps/cpu", "\xb1\x01\a"},
	{"crypto/internal/fips140deps/godebug", "\xb9\x01"},
	{"crypto/internal/fips140deps/time", "\xc9\x02"},
	{"crypto/internal/fips140hash", "7\x1c3\xc9\x01"},
	{"crypto/internal/fips140only", ")\r\x01\x01N3<"},
	{"crypto/internal/fips140test", ""},
	{"crypto/internal/hpke", "\x0e\x01\x01\x03\x056#+hM"},
	{"crypto/internal/impl", "\xb9\x02"},
	{"crypto/internal/randutil", "\xf5\x01\x12"},
	{"crypto/internal/sysrand", "qo! \r\r\x01\x01\f\x06"},
	{"crypto/internal/sysrand/internal/seccomp", "q"},
	{"crypto/md5", "\x0e6-\x15\x16h"},
	{"crypto/mlkem", "1"},
	{"crypto/pbkdf2", "4\x0f\x01-\x15"},
	{"crypto/rand", "\x1a\b\a\x1b\x04\x01(\x84\x01\rM"},
	{"crypto/rc4", "%\x1f-\xc7\x01"},
	{"crypto/rsa", "\x0e\f\x01\v\x0f\x0e\x01\x04\x06\a\x1c\x03\x123<\f\x01"},
	{"crypto/sha1", "\x0e\f*\x03*\x15\x16\x15S"},
	{"crypto/sha256", "\x0e\f\x1cP"},
	{"crypto/sha3", "\x0e)O\xc9\x01"},
	{"crypto/sha512", "\x0e\f\x1eN"},
	{"crypto/subtle", "\x1e\x1c\x9c\x01X"},
	{"crypto/tls", "\x03\b\x02\x01\x01\x01\x01\x02\x01\x01\x01\x02\x01\x01\t\x01\r\n\x01\n\x05\x03\x01\x01\x01\x01\x02\x01\x02\x01\x17\x02\x03\x12\x16\x15\b<\x16\x16\r\b\x01\x01\x01\x02\x01\r\x06\x02\x01\x0f"},
	{"crypto/tls/internal/fips140tls", "\x17\xa5\x02"},
	{"crypto/x509", "\x03\v\x01\x01\x01\x01\x01\x01\x01\x015\x05\x01\x01\x02\x05\x0e\x06\x02\x02\x03E\x039\x01\x02\b\x01\x01\x02\a\x10\x05\x01\x06\x02\x05\b\x02\x01\x02\x0e\x02\x01\x01\x02\x03\x01"},
	{"crypto/x509/pkix", "g\x06\a\x8e\x01G"},
	{"database/sql", "\x03\nN\x16\x03\x81\x01\v\a\"\x05\b\x02\x03\x01\r\x02\x02\x02"},
	{"database/sql/driver", "\rd\x03\xb5\x01\x0f\x11"},
	{"debug/buildinfo", "\x03[\x02\x01\x01\b\a\x03e\x1a\x02\x01+\x0f\x1f"},
	{"debug/dwarf", "\x03g\a\x03\x81\x011\x11\x01\x01"},
	{"debug/elf", "\x03\x06T\r\a\x03e\x1b\x01\f \x17\x01\x16"},
	{"debug/gosym", "\x03g\n\xc3\x01\x01\x01\x02"},
	{"debug/macho", "\x03\x06T\r\ne\x1c,\x17\x01"},
	{"debug/pe", "\x03\x06T\r\a\x03e\x1c,\x17\x01\x16"},
	{"debug/plan9obj", "j\a\x03e\x1c,"},
	{"embed", "q*A\x19\x01S"},
	{"embed/internal/embedtest", ""},
	{"encoding", ""},
	{"encoding/ascii85", "\xf5\x01C"},
	{"encoding/asn1", "\x03n\x03e(\x01'\r\x02\x01\x10\x03\x01"},
	{"encoding/base32", "\xf5\x01A\x02"},
	{"encoding/base64", "\x9c\x01YA\x02"},
	{"encoding/binary", "q\x84\x01\f(\r\x05"},
	{"encoding/csv", "\x02\x01n\x03\x81\x01D\x12\x02"},
	{"encoding/gob", "\x02c\x05\a\x03e\x1c\v\x01\x03\x1d\b\x12\x01\x0f\x02"},
	{"encoding/hex", "q\x03\x81\x01A\x03"},
	{"encoding/json", "\x03\x01a\x04\b\x03\x81\x01\f(\r\x02\x01\x02\x10\x01\x01\x02"},
	{"encoding/pem", "\x03f\b\x84\x01A\x03"},
	{"encoding/xml", "\x02\x01b\f\x03\x81\x014\x05\n\x01\x02\x10\x02"},
	{"errors", "\xcc\x01\x83\x01"},
	{"expvar", "nK@\b\v\x15\r\b\x02\x03\x01\x11"},
	{"flag", "e\f\x03\x81\x01,\b\x05\b\x02\x01\x10"},
	{"fmt", "qE&\x19\f \b\r\x02\x03\x12"},
	{"go/ast", "\x03\x01p\x0e\x01r\x03)\b\r\x02\x01\x12\x02"},
	{"go/build", "\x02\x01n\x03\x01\x02\x02\a\x02\x01\x17\x1f\x04\x02\b\x1b\x13\x01+\x01\x04\x01\a\b\x02\x01\x12\x02\x02"},
	{"go/build/constraint", "q\xc7\x01\x01\x12\x02"},
	{"go/constant", "t\x0f~\x01\x024\x01\x02\x12"},
	{"go/doc", "\x04p\x01\x05\t=51\x10\x02\x01\x12\x02"},
	{"go/doc/comment", "\x03q\xc2\x01\x01\x01\x01\x12\x02"},
	{"go/format", "\x03q\x01\v\x01\x02rD"},
	{"go/importer", "v\a\x01\x01\x04\x01q9"},
	{"go/internal/gccgoimporter", "\x02\x01[\x13\x03\x04\v\x01o\x02,\x01\x05\x11\x01\f\b"},
	{"go/internal/gcimporter", "\x02r\x0f\x010\x05\r/,\x15\x03\x02"},
	{"go/internal/srcimporter", "t\x01\x01\n\x03\x01q,\x01\x05\x12\x02\x14"},
	{"go/parser", "\x03n\x03\x01\x02\v\x01r\x01+\x06\x12"},
	{"go/printer", "t\x01\x02\x03\tr\f \x15\x02\x01\x02\v\x05\x02"},
	{"go/scanner", "\x03q\x0fr2\x10\x01\x13\x02"},
	{"go/token", "\x04p\x84\x01>\x02\x03\x01\x0f\x02"},
	{"go/types", "\x03\x01\x06g\x03\x01\x03\b\x03\x024\x062\x04\x03\t \x06\a\b\x01\x01\x01\x02\x01\x0f\x02\x02"},
	{"go/version", "\xbe\x01{"},
	{"hash", "\xf5\x01"},
	{"hash/adler32", "q\x15\x16"},
	{"hash/crc32", "q\x15\x16\x15\x8a\x01\x01\x13"},
	{"hash/crc64", "q\x15\x16\x9f\x01"},
	{"hash/fnv", "q\x15\x16h"},
	{"hash/maphash", "\x86\x01\x11<|"},
	{"html", "\xb9\x02\x02\x12"},
	{"html/template", "\x03k\x06\x18-<\x01\n!\x05\x01\x02\x03\f\x01\x02\f\x01\x03\x02"},
	{"image", "\x02o\x1ef\x0f4\x03\x01"},
	{"image/color", ""},
	{"image/color/palette", "\x8f\x01"},
	{"image/draw", "\x8e\x01\x01\x04"},
	{"image/gif", "\x02\x01\x05i\x03\x1a\x01\x01\x01\vY"},
	{"image/internal/imageutil", "\x8e\x01"},
	{"image/jpeg", "\x02o\x1d\x01\x04b"},
	{"image/png", "\x02\aa\n\x12\x02\x06\x01fC"},
	{"index/suffixarray", "\x03g\a\x84\x01\f+\n\x01"},
	{"internal/abi", "\xb8\x01\x97\x01"},
	{"internal/asan", "\xcf\x02"},
	{"internal/bisect", "\xae\x02\r\x01"},
	{"internal/buildcfg", "tGf\x06\x02\x05\n\x01"},
	{"internal/bytealg", "\xb1\x01\x9e\x01"},
	{"internal/byteorder", ""},
	{"internal/cfg", ""},
	{"internal/cgrouptest", "tZS\x06\x0f\x02\x01\x04\x01"},
	{"internal/chacha8rand", "\x9c\x01\x15\a\x97\x01"},
	{"internal/copyright", ""},
	{"internal/coverage", ""},
	{"internal/coverage/calloc", ""},
	{"internal/coverage/cfile", "n\x06\x16\x17\x01\x02\x01\x01\x01\x01\x01\x01\x01\"\x02&,\x06\a\n\x01\x03\r\x06"},
	{"internal/coverage/cformat", "\x04p-\x04P\v6\x01\x02\r"},
	{"internal/coverage/cmerge", "t-`"},
	{"internal/coverage/decodecounter", "j\n-\v\x02G,\x17\x17"},
	{"internal/coverage/decodemeta", "\x02h\n\x16\x17\v\x02G,"},
	{"internal/coverage/encodecounter", "\x02h\n-\f\x01\x02E\v!\x15"},
	{"internal/coverage/encodemeta", "\x02\x01g\n\x12\x04\x17\r\x02E,."},
	{"internal/coverage/pods", "\x04p-\x80\x01\x06\x05\n\x02\x01"},
	{"internal/coverage/rtcov", "\xcf\x02"},
	{"internal/coverage/slicereader", "j\n\x81\x01Z"},
	{"internal/coverage/slicewriter", "t\x81\x01"},
	{"internal/coverage/stringtab", "t8\x04E"},
	{"internal/coverage/test", ""},
	{"internal/coverage/uleb128", ""},
	{"internal/cpu", "\xcf\x02"},
	{"internal/dag", "\x04p\xc2\x01\x03"},
	{"internal/diff", "\x03q\xc3\x01\x02"},
	{"internal/exportdata", "\x02\x01n\x03\x02c\x1c,\x01\x05\x11\x01\x02"},
	{"internal/filepathlite", "q*A\x1a@"},
	{"internal/fmtsort", "\x04\xa5\x02\r"},
	{"internal/fuzz", "\x03\nE\x18\x04\x03\x03\x01\v\x036<\f\x03\x1d\x01\x05\x02\x05\n\x01\x02\x01\x01\f\x04\x02"},
	{"internal/goarch", ""},
	{"internal/godebug", "\x99\x01!\x81\x01\x01\x13"},
	{"internal/godebugs", ""},
	{"internal/goexperiment", ""},
	{"internal/goos", ""},
	{"internal/goroot", "\xa1\x02\x01\x05\x12\x02"},
	{"internal/gover", "\x04"},
	{"internal/goversion", ""},
	{"internal/lazyregexp", "\xa1\x02\v\r\x02"},
	{"internal/lazytemplate", "\xf5\x01,\x18\x02\f"},
	{"internal/msan", "\xcf\x02"},
	{"internal/nettrace", ""},
	{"internal/obscuretestdata", "i\x8c\x01,"},
	{"internal/oserror", "q"},
	{"internal/pkgbits", "\x03O\x18\a\x03\x04\vr\r\x1f\r\n\x01"},
	{"internal/platform", ""},
	{"internal/poll", "qj\x05\x159\r\x01\x01\f\x06"},
	{"internal/profile", "\x03\x04j\x03\x81\x017\n\x01\x01\x01\x10"},
	{"internal/profilerecord", ""},
	{"internal/race", "\x97\x01\xb8\x01"},
	{"internal/reflectlite", "\x97\x01!:<!"},
	{"internal/runtime/atomic", "\xb8\x01\x97\x01"},
	{"internal/runtime/cgroup", "\x9b\x01<\x04t"},
	{"internal/runtime/exithook", "\xcd\x01\x82\x01"},
	{"internal/runtime/gc", "\xb8\x01"},
	{"internal/runtime/gc/internal/gen", "\n`\n\x17j\x04\v\x1d\b\x10\x02"},
	{"internal/runtime/gc/scan", "\xb1\x01\a\x18\x06y"},
	{"internal/runtime/maps", "\x97\x01\x01 \n\t\t\x02y"},
	{"internal/runtime/math", "\xb8\x01"},
	{"internal/runtime/startlinetest", ""},
	{"internal/runtime/sys", "\xb8\x01\x04"},
	{"internal/runtime/syscall/linux", "\xb8\x01\x97\x01"},
	{"internal/runtime/wasitest", ""},
	{"internal/saferio", "\xf5\x01Z"},
	{"internal/singleflight", "\xbb\x02"},
	{"internal/strconv", "\x84\x02K"},
	{"internal/stringslite", "\x9b\x01\xb4\x01"},
	{"internal/sync", "\x97\x01!\x13q\x13"},
	{"internal/synctest", "\x97\x01\xb8\x01"},
	{"internal/syscall/execenv", "\xbd\x02"},
	{"internal/syscall/unix", "\xae\x02\x0e\x01\x12"},
	{"internal/sysinfo", "\x02\x01\xae\x01D,\x18\x02"},
	{"internal/syslist", ""},
	{"internal/testenv", "\x03\nd\x02\x01)\x1b\x0f/+\x01\x05\a\n\x01\x02\x02\x01\v"},
	{"internal/testhash", "\x03\x83\x01o\x118\v"},
	{"internal/testlog", "\xbb\x02\x01\x13"},
	{"internal/testpty", "q\x03\xad\x01"},
	{"internal/trace", "\x02\x01\x01\x06`\a\x03u\x03\x03\x06\x03\t5\x01\x01\x01\x10\x06"},
	{"internal/trace/internal/testgen", "\x03g\ns\x03\x02\x03\x011\v\r\x10"},
	{"internal/trace/internal/tracev1", "\x03\x01f\a\x03{\x06\f5\x01"},
	{"internal/trace/raw", "\x02h\nx\x03\x06C\x01\x12"},
	{"internal/trace/testtrace", "\x02\x01n\x03o\x04\x03\x05\x01\x05,\v\x02\b\x02\x01\x05"},
	{"internal/trace/tracev2", ""},
	{"internal/trace/traceviewer", "\x02a\v\x06\x19<\x1e\a\a\x04\b\v\x15\x01\x05\a\n\x01\x02\x0e"},
	{"internal/trace/traceviewer/format", ""},
	{"internal/trace/version", "tx\t"},
	{"internal/txtar", "\x03q\xad\x01\x18"},
	{"internal/types/errors", "\xb8\x02"},
	{"internal/unsafeheader", "\xcf\x02"},
	{"internal/xcoff", "]\r\a\x03e\x1c,\x17\x01"},
	{"internal/zstd", "j\a\x03\x81\x01\x0f"},
	{"io", "q\xca\x01"},
	{"io/fs", "q**01\x10\x13\x04"},
	{"io/ioutil", "\xf5\x01\x01+\x15\x03"},
	{"iter", "\xcb\x01c!"},
	{"log", "t\x81\x01\x05'\r\r\x01\r"},
	{"log/internal", ""},
	{"log/slog", "\x03\nX\t\x03\x03\x81\x01\x04\x01\x02\x02\x03(\x05\b\x02\x01\x02\x01\r\x02\x02\x02"},
	{"log/slog/internal", ""},
	{"log/slog/internal/benchmarks", "\rd\x03\x81\x01\x06\x03:\x11"},
	{"log/slog/internal/buffer", "\xbb\x02"},
	{"log/syslog", "q\x03\x85\x01\x12\x16\x18\x02\x0e"},
	{"maps", "\xf8\x01W"},
	{"math", "\xb1\x01SK"},
	{"math/big", "\x03n\x03(\x15D\f\x03\x020\x02\x01\x02\x14"},
	{"math/big/internal/asmgen", "\x03\x01p\x90\x012\x03"},
	{"math/bits", "\xcf\x02"},
	{"math/cmplx", "\x81\x02\x03"},
	{"math/rand", "\xb9\x01H:\x01\x13"},
	{"math/rand/v2", "q+\x03b\x03K"},
	{"mime", "\x02\x01f\b\x03\x81\x01\v!\x15\x03\x02\x10\x02"},
	{"mime/multipart", "\x02\x01K#\x03E<\v\x01\a\x02\x15\x02\x06\x0f\x02\x01\x16"},
	{"mime/quotedprintable", "\x02\x01q\x81\x01"},
	{"net", "\x04\td*\x1e\n\x05\x12\x01\x01\x04\x15\x01%\x06\r\b\x05\x01\x01\f\x06\a"},
	{"net/http", "\x02\x01\x03\x01\x04\x02A\b\x13\x01\a\x03E<\x01\x03\a\x01\x03\x02\x02\x01\x02\x06\x02\x01\x01\n\x01\x01\x05\x01\x02\x05\b\x01\x01\x01\x02\x01\r\x02\x02\x02\b\x01\x01\x01"},
	{"net/http/cgi", "\x02T\x1b\x03\x81\x01\x04\a\v\x01\x13\x01\x01\x01\x04\x01\x05\x02\b\x02\x01\x10\x0e"},
	{"net/http/cookiejar", "\x04m\x03\x97\x01\x01\b\f\x16\x03\x02\x0e\x04"},
	{"net/http/fcgi", "\x02\x01\n]\a\x03\x81\x01\x16\x01\x01\x14\x18\x02\x0e"},
	{"net/http/httptest", "\x02\x01\nI\x02\x1b\x01\x81\x01\x04\x12\x01\n\t\x02\x17\x01\x02\x0e\x0e"},
	{"net/http/httptrace", "\rImH\x14\n "},
	{"net/http/httputil", "\x02\x01\nd\x03\x81\x01\x04\x0f\x03\x01\x05\x02\x01\v\x01\x19\x02\x0e\x0e"},
	{"net/http/internal", "\x02\x01n\x03\x81\x01"},
	{"net/http/internal/ascii", "\xb9\x02\x12"},
	{"net/http/internal/httpcommon", "\rd\x03\x9d\x01\x0e\x01\x17\x01\x01\x02\x1c\x02"},
	{"net/http/internal/testcert", "\xb9\x02"},
	{"net/http/pprof", "\x02\x01\ng\x18-\x02\x0e,\x04\x13\x14\x01\r\x04\x03\x01\x02\x01\x10"},
	{"net/internal/cgotest", ""},
	{"net/internal/socktest", "t\xc7\x01\x02"},
	{"net/mail", "\x02o\x03\x81\x01\x04\x0f\x03\x14\x1a\x02\x0e\x04"},
	{"net/netip", "\x04m*\x01e\x034\x16"},
	{"net/rpc", "\x02j\x05\x03\x0f\nh\x04\x12\x01\x1d\r\x03\x02"},
	{"net/rpc/jsonrpc", "n\x03\x03\x81\x01\x16\x11\x1f"},
	{"net/smtp", "\x192\v\x13\b\x03\x81\x01\x16\x14\x1a"},
	{"net/textproto", "\x02\x01n\x03\x81\x01\f\n-\x01\x02\x14"},
	{"net/url", "q\x03\xa7\x01\v\x10\x02\x01\x16"},
	{"os", "q*\x01\x19\x03\x10\x13\x01\x03\x01\x05\x10\x018\b\x05\x01\x01\f\x06"},
	{"os/exec", "\x03\ndH&\x01\x15\x01+\x06\a\n\x01\x04\f"},
	{"os/exec/internal/fdtest", "\xbd\x02"},
	{"os/signal", "\r\x94\x02\x15\x05\x02"},
	{"os/user", "\x02\x01n\x03\x81\x01,\r\n\x01\x02"},
	{"path", "q*\xb2\x01"},
	{"path/filepath", "q*\x1aA+\r\b\x03\x04\x10"},
	{"plugin", "q"},
	{"reflect", "q&\x04\x1d\x13\b\x03\x05\x17\x06\t-\n\x03\x10\x02\x02"},
	{"reflect/internal/example1", ""},
	{"reflect/internal/example2", ""},
	{"regexp", "\x03\xf2\x018\t\x02\x01\x02\x10\x02"},
	{"regexp/syntax", "\xb6\x02\x01\x01\x01\x02\x10\x02"},
	{"runtime", "\x97\x01\x04\x01\x03\f\x06\a\x02\x01\x01\x0e\x03\x01\x01\x01\x02\x01\x01\x02\x01\x04\x01\x10c"},
	{"runtime/coverage", "\xa3\x01R"},
	{"runtime/debug", "tTY\r\b\x02\x01\x10\x06"},
	{"runtime/metrics", "\xba\x01G-!"},
	{"runtime/pprof", "\x02\x01\x01\x03\x06]\a\x03#$\x0f+\f \r\b\x01\x01\x01\x02\x02\t\x03\x06"},
	{"runtime/race", "\xb4\x02"},
	{"runtime/race/internal/amd64v1", ""},
	{"runtime/trace", "\rd\x03x\t9\b\x05\x01\r\x06"},
	{"slices", "\x04\xf4\x01\fK"},
	{"sort", "\xcc\x0182"},
	{"strconv", "q*@\x01q"},
	{"strings", "q&\x04A\x19\x03\f7\x10\x02\x02"},
	{"structs", ""},
	{"sync", "\xcb\x01\x12\x01P\x0e\x13"},
	{"sync/atomic", "\xcf\x02"},
	{"syscall", "q'\x03\x01\x1c\n\x03\x06\f\x04S\b\x05\x01\x13"},
	{"testing", "\x03\nd\x02\x01W\x16\x14\f\x05\x1b\x06\x02\x05\x02\x05\x01\x02\x01\x02\x01\r\x02\x04"},
	{"testing/fstest", "q\x03\x81\x01\x01\n&\x10\x03\b\b"},
	{"testing/internal/testdeps", "\x02\v\xaa\x01.\x10,\x03\x05\x03\x06\a\x02\x0e"},
	{"testing/iotest", "\x03n\x03\x81\x01\x04"},
	{"testing/quick", "s\x01\x8d\x01\x05#\x10\x10"},
	{"testing/slogtest", "\rd\x03\x87\x01.\x05\x10\v"},
	{"testing/synctest", "\xde\x01`\x11"},
	{"text/scanner", "\x03q\x81\x01,*\x02"},
	{"text/tabwriter", "t\x81\x01X"},
	{"text/template", "q\x03B?\x01\n \x01\x05\x01\x02\x05\v\x02\r\x03\x02"},
	{"text/template/parse", "\x03q\xba\x01\n\x01\x12\x02"},
	{"time", "q*\x1e#(*\r\x02\x12"},
	{"time/tzdata", "q\xcc\x01\x12"},
	{"unicode", ""},
	{"unicode/utf16", ""},
	{"unicode/utf8", ""},
	{"unique", "\x97\x01!$\x01Q\r\x01\x13\x12"},
	{"unsafe", ""},
	{"vendor/golang.org/x/crypto/chacha20", "\x10Z\a\x93\x01*&"},
	{"vendor/golang.org/x/crypto/chacha20poly1305", "\x10Z\a\xdf\x01\x04\x01\a"},
	{"vendor/golang.org/x/crypto/cryptobyte", "g\n\x03\x8e\x01' \n"},
	{"vendor/golang.org/x/crypto/cryptobyte/asn1", ""},
	{"vendor/golang.org/x/crypto/internal/alias", "\xcf\x02"},
	{"vendor/golang.org/x/crypto/internal/poly1305", "U\x15\x9a\x01"},
	{"vendor/golang.org/x/net/dns/dnsmessage", "q"},
	{"vendor/golang.org/x/net/http/httpguts", "\x8b\x02\x14\x1a\x14\r"},
	{"vendor/golang.org/x/net/http/httpproxy", "q\x03\x97\x01\x10\x05\x01\x18\x14\r"},
	{"vendor/golang.org/x/net/http2/hpack", "\x03n\x03\x81\x01F"},
	{"vendor/golang.org/x/net/idna", "t\x8d\x018\x14\x10\x02\x01"},
	{"vendor/golang.org/x/net/nettest", "\x03g\a\x03\x81\x01\x11\x05\x16\x01\f\n\x01\x02\x02\x01\v"},
	{"vendor/golang.org/x/sys/cpu", "\xa1\x02\r\n\x01\x16"},
	{"vendor/golang.org/x/text/secure/bidirule", "q\xdc\x01\x11\x01"},
	{"vendor/golang.org/x/text/transform", "\x03n\x84\x01X"},
	{"vendor/golang.org/x/text/unicode/bidi", "\x03\bi\x85\x01>\x16"},
	{"vendor/golang.org/x/text/unicode/norm", "j\n\x81\x01F\x12\x11"},
	{"weak", "\x97\x01\x97\x01!"},
}

// bootstrap is the list of bootstrap packages extracted from cmd/dist.
var bootstrap = map[string]bool{
	"cmp":                                     true,
	"cmd/asm":                                 true,
	"cmd/asm/internal/arch":                   true,
	"cmd/asm/internal/asm":                    true,
	"cmd/asm/internal/flags":                  true,
	"cmd/asm/internal/lex":                    true,
	"cmd/cgo":                                 true,
	"cmd/compile":                             true,
	"cmd/compile/internal/abi":                true,
	"cmd/compile/internal/abt":                true,
	"cmd/compile/internal/amd64":              true,
	"cmd/compile/internal/arm":                true,
	"cmd/compile/internal/arm64":              true,
	"cmd/compile/internal/base":               true,
	"cmd/compile/internal/bitvec":             true,
	"cmd/compile/internal/compare":            true,
	"cmd/compile/internal/coverage":           true,
	"cmd/compile/internal/deadlocals":         true,
	"cmd/compile/internal/devirtualize":       true,
	"cmd/compile/internal/dwarfgen":           true,
	"cmd/compile/internal/escape":             true,
	"cmd/compile/internal/gc":                 true,
	"cmd/compile/internal/importer":           true,
	"cmd/compile/internal/inline":             true,
	"cmd/compile/internal/inline/inlheur":     true,
	"cmd/compile/internal/inline/interleaved": true,
	"cmd/compile/internal/ir":                 true,
	"cmd/compile/internal/liveness":           true,
	"cmd/compile/internal/logopt":             true,
	"cmd/compile/internal/loong64":            true,
	"cmd/compile/internal/loopvar":            true,
	"cmd/compile/internal/mips":               true,
	"cmd/compile/internal/mips64":             true,
	"cmd/compile/internal/noder":              true,
	"cmd/compile/internal/objw":               true,
	"cmd/compile/internal/pgoir":              true,
	"cmd/compile/internal/pkginit":            true,
	"cmd/compile/internal/ppc64":              true,
	"cmd/compile/internal/rangefunc":          true,
	"cmd/compile/internal/reflectdata":        true,
	"cmd/compile/internal/riscv64":            true,
	"cmd/compile/internal/rttype":             true,
	"cmd/compile/internal/s390x":              true,
	"cmd/compile/internal/ssa":                true,
	"cmd/compile/internal/ssagen":             true,
	"cmd/compile/internal/staticdata":         true,
	"cmd/compile/internal/staticinit":         true,
	"cmd/compile/internal/syntax":             true,
	"cmd/compile/internal/test":               true,
	"cmd/compile/internal/typebits":           true,
	"cmd/compile/internal/typecheck":          true,
	"cmd/compile/internal/types":              true,
	"cmd/compile/internal/types2":             true,
	"cmd/compile/internal/walk":               true,
	"cmd/compile/internal/wasm":               true,
	"cmd/compile/internal/x86":                true,
	"cmd/internal/archive":                    true,
	"cmd/internal/bio":                        true,
	"cmd/internal/codesign":                   true,
	"cmd/internal/dwarf":                      true,
	"cmd/internal/edit":                       true,
	"cmd/internal/gcprog":                     true,
	"cmd/internal/goobj":                      true,
	"cmd/internal/hash":                       true,
	"cmd/internal/macho":                      true,
	"cmd/internal/obj":                        true,
	"cmd/internal/obj/arm":                    true,
	"cmd/internal/obj/arm64":                  true,
	"cmd/internal/obj/loong64":                true,
	"cmd/internal/obj/mips":                   true,
	"cmd/internal/obj/ppc64":                  true,
	"cmd/internal/obj/riscv":                  true,
	"cmd/internal/obj/s390x":                  true,
	"cmd/internal/obj/wasm":                   true,
	"cmd/internal/obj/x86":                    true,
	"cmd/internal/objabi":                     true,
	"cmd/internal/par":                        true,
	"cmd/internal/pgo":                        true,
	"cmd/internal/pkgpath":                    true,
	"cmd/internal/quoted":                     true,
	"cmd/internal/src":                        true,
	"cmd/internal/sys":                        true,
	"cmd/internal/telemetry":                  true,
	"cmd/internal/telemetry/counter":          true,
	"cmd/link":                                true,
	"cmd/link/internal/amd64":                 true,
	"cmd/link/internal/arm":                   true,
	"cmd/link/internal/arm64":                 true,
	"cmd/link/internal/benchmark":             true,
	"cmd/link/internal/dwtest":                true,
	"cmd/link/internal/ld":                    true,
	"cmd/link/internal/loadelf":               true,
	"cmd/link/internal/loader":                true,
	"cmd/link/internal/loadmacho":             true,
	"cmd/link/internal/loadpe":                true,
	"cmd/link/internal/loadxcoff":             true,
	"cmd/link/internal/loong64":               true,
	"cmd/link/internal/mips":                  true,
	"cmd/link/internal/mips64":                true,
	"cmd/link/internal/ppc64":                 true,
	"cmd/link/internal/riscv64":               true,
	"cmd/link/internal/s390x":                 true,
	"cmd/link/internal/sym":                   true,
	"cmd/link/internal/wasm":                  true,
	"cmd/link/internal/x86":                   true,
	"compress/flate":                          true,
	"compress/zlib":                           true,
	"container/heap":                          true,
	"debug/dwarf":                             true,
	"debug/elf":                               true,
	"debug/macho":                             true,
	"debug/pe":                                true,
	"go/build/constraint":                     true,
	"go/constant":                             true,
	"go/version":                              true,
	"internal/abi":                            true,
	"internal/coverage":                       true,
	"cmd/internal/cov/covcmd":                 true,
	"internal/bisect":                         true,
	"internal/buildcfg":                       true,
	"internal/exportdata":                     true,
	"internal/goarch":                         true,
	"internal/godebugs":                       true,
	"internal/goexperiment":                   true,
	"internal/goroot":                         true,
	"internal/gover":                          true,
	"internal/goversion":                      true,
	"internal/lazyregexp":                     true,
	"internal/pkgbits":                        true,
	"internal/platform":                       true,
	"internal/profile":                        true,
	"internal/race":                           true,
	"internal/runtime/gc":                     true,
	"internal/saferio":                        true,
	"internal/syscall/unix":                   true,
	"internal/types/errors":                   true,
	"internal/unsafeheader":                   true,
	"internal/xcoff":                          true,
	"internal/zstd":                           true,
	"math/bits":                               true,
	"sort":                                    true,
}

// BootstrapVersion is the minor version of Go used during toolchain
// bootstrapping. Packages for which [IsBootstrapPackage] must not use
// features of Go newer than this version.
const BootstrapVersion = Version(24) // go1.24.6
